H1 {
    text-align: center;
    color: #CC0000;
}
H2, H3 {
    color: #CC0000;
}

img {
    border: 0;
}

H1, H2, H3,  div, td, a, span, p, body {
    font-family: Verdana, Geneva, Arial, Helvetica, sans-serif;
}

body, p, a {
    font-size: 14px;
}

CAPTION { font-family: Verdana, Geneva, Arial, Helvetica, sans-serif; font-weight: bold }

DIV.qindex {
    width: 100%;
    background-color: #FFFBD7;
    border: 1px solid #CC0000;
    text-align: center;
    margin: 2px;
    padding: 2px;
    line-height: 120%;
}

A.qindex {
    text-decoration: none;
    font-weight: bold;
    color: #CC0000;
    padding: 2px;
}
A.qindex:visited {
    text-decoration: none;
    font-weight: bold;
    color: #CC0000;
    padding: 2px;
}
A.qindex:hover {
    text-decoration: none;
    background-color: #EDCF88;
    padding: 2px;
}
A.qindexHL {
    text-decoration: none;
    font-weight: bold;
    background-color: #CC0000;
    color: #ffffff;
    padding: 2px 6px;
    border: 1px double #CC0000;
}
A.qindexHL:hover {
    text-decoration: none;
    background-color: #CC0000;
    color: #ffffff;
    padding: 2px 6px;
}
A.qindexHL:visited { text-decoration: none; background-color: #CC0000; color: #ffffff }

A.el {
    text-decoration: none;
    font-weight: bold;
}
A.elRef { font-weight: bold }
A.code { text-decoration: none; font-weight: normal; color: #1A419D}
A.codeRef { font-weight: normal; color: #1A419D}

A:hover { text-decoration: none; background-color: #EDCF88 }

DL.el { margin-left: -1cm }
PRE.fragment {
    border: 1px solid #CCCCCC;
    background-color: #f5f5f5;
    margin-top: 4px;
    margin-bottom: 4px;
    margin-left: 2px;
    margin-right: 8px;
    padding-left: 6px;
    padding-right: 6px;
    padding-top: 4px;
    padding-bottom: 4px;
}
DIV.fragment {
    border: 1px solid #CCCCCC;
    background-color: #f5f5f5;
    padding: 6px;
}
DIV.ah { background-color: black; font-weight: bold; color: #ffffff; margin-bottom: 3px; margin-top: 3px }
TD.md { background-color: #FFFBD7; font-weight: bold; color: #666666; }
TD.mdname1 { background-color: #FFFBD7; font-weight: bold; color: #602020; }
TD.mdname { background-color: #FFFBD7; font-weight: bold; color: #602020; width: 600px; }
DIV.groupHeader {
    margin-left: 16px;
    margin-top: 12px;
    margin-bottom: 6px;
    font-weight: bold;
    font-family: Verdana, Geneva, Arial, Helvetica, sans-serif;
}
DIV.groupText { margin-left: 16px; font-style: italic; font-size: 14px }
BODY {
    background: white;
    color: black;
    margin-right: 20px;
    margin-left: 20px;
}
TD.indexkey {
    background-color: #FFFBD7;
    font-weight: bold;
    padding-right  : 10px;
    padding-top    : 2px;
    padding-left   : 10px;
    padding-bottom : 2px;
    margin-left    : 0px;
    margin-right   : 0px;
    margin-top     : 2px;
    margin-bottom  : 2px;
    border: 1px solid #CCCCCC;
}
TD.indexvalue {
    background-color: #FFFBD7;
    font-style: italic;
    padding-right  : 10px;
    padding-top    : 2px;
    padding-left   : 10px;
    padding-bottom : 2px;
    margin-left    : 0px;
    margin-right   : 0px;
    margin-top     : 2px;
    margin-bottom  : 2px;
    border: 1px solid #CCCCCC;
}
TR.memlist {
   background-color: #f0f0f0;
}
P.formulaDsp { text-align: center; }
IMG.formulaDsp { }
IMG.formulaInl { vertical-align: middle; }
SPAN.keyword       { color: #008000 }
SPAN.keywordtype   { color: #604020 }
SPAN.keywordflow   { color: #e08000 }
SPAN.comment       { color: #800000 }
SPAN.preprocessor  { color: #806020 }
SPAN.stringliteral { color: #002080 }
SPAN.charliteral   { color: #008080 }

.mdTable {
    border: 0px solid #868686;
    background-color: #FFFBD7;
}
#F4F4FB

.mdRow {
    padding: 8px 10px;
}
.mdescLeft {
    padding: 0px 8px 4px 8px;
    font-size: 14px;
    font-style: italic;
    background-color: #FFFBD7;
    border-top: 1px none #E0E0E0;
    border-right: 1px none #E0E0E0;
    border-bottom: 1px none #E0E0E0;
    border-left: 1px none #E0E0E0;
    margin: 0px;
}
.mdescRight {
    padding: 0px 8px 4px 8px;
    font-size: 14px;
    font-style: italic;
    background-color: #FFFBD7;
    border-top: 1px none #E0E0E0;
    border-right: 1px none #E0E0E0;
    border-bottom: 1px none #E0E0E0;
    border-left: 1px none #E0E0E0;
    margin: 0px;
}
.memItemLeft {
    padding: 1px 0px 0px 8px;
    margin: 4px;
    border-top-width: 1px;
    border-right-width: 1px;
    border-bottom-width: 1px;
    border-left-width: 1px;
    border-top-style: solid;
    border-top-color: #E0E0E0;
    border-right-color: #E0E0E0;
    border-bottom-color: #E0E0E0;
    border-left-color: #E0E0E0;
    border-right-style: none;
    border-bottom-style: none;
    border-left-style: none;
    background-color: #FFFBD7;
    font-family: Verdana, Geneva, Arial, Helvetica, sans-serif;
    font-size: 12px;
}
.memItemRight {
    padding: 1px 8px 0px 8px;
    margin: 4px;
    border-top-width: 1px;
    border-right-width: 1px;
    border-bottom-width: 1px;
    border-left-width: 1px;
    border-top-style: solid;
    border-top-color: #E0E0E0;
    border-right-color: #E0E0E0;
    border-bottom-color: #E0E0E0;
    border-left-color: #E0E0E0;
    border-right-style: none;
    border-bottom-style: none;
    border-left-style: none;
    background-color: #FFFBD7;
    font-family: Verdana, Geneva, Arial, Helvetica, sans-serif;
    font-size: 13px;
}
.memitem {
    padding: 0px 8px 4px 8px;
    font-size: 14px;
    background-color: #FFFBD7;
    border-top: 1px none #E0E0E0;
    border-right: 1px none #E0E0E0;
    border-bottom: 1px none #E0E0E0;
    border-left: 1px none #E0E0E0;
    margin: 0px;
}
.search     { color: #003399;
              font-weight: bold;
}
FORM.search {
    margin-bottom: 0px;
    margin-top: 0px;
}
INPUT.search {
    font-size: 75%;
    color: #000080;
    font-weight: normal;
    background-color: #eeeeff;
}
TD.tiny      { font-size: 75%;}
a {
    color: #000000;
}
a:visited {
    color: #000000;
}

A.logo1 {
    padding: 2px;
}
A.logo1:visited {
    padding: 2px;
}
A.logo1:hover {
    background-color: #FFFBD7;
    padding: 2px;
}

A.logo2 {
    padding: 2px;
}
A.logo2:visited {
    padding: 2px;
}
A.logo2:hover {
    background-color: #FFFFFF;
    padding: 2px;
}

.small {
    font-family: Verdana, Geneva, Arial, Helvetica, sans-serif;
    font-size: 9;
}
